/***************************************************************/
/* tea (Tag Extraction and Annotation)                         */
/*                                                             */
/*  Usage: tea {-a|-e} index_file doc_file                     */
/*                                                             */
/*    Annotate tag to documents or extract tag from documents. */
/*    See README file for more detail.                         */
/*                                                             */
/*  Option:                                                    */
/*     e : Extraction mode (Output to index file)              */
/*     a : Annotation mode (Output to stdout)                  */
/*                                                             */
/*                                                             */
/*                            Initial   November, 1, 1997      */
/*                            Latest    August, 5, 1999        */
/*                            Author    Satoshi Sekine (NYU)   */
/***************************************************************/

/********************/
/* Include files    */
/********************/
#include <stdio.h>
#include <ctype.h>


/********************/
/* Macro variables  */
/********************/
#define DEBUG              1


#define MAX_DOCNO      10000       /* Number of documents treated by the system */
#define MAX_DOC_LEN    50000       /* Length of document */

#define MAX_TAGSET       100       /* Number of possible tagset */
#define MAX_TAG_STRING   500       /* Length of tag string */
#define MAX_DOCNO_STRING  50       /* Length of document number string */
#define MAX_WORD_STRING 5000       /* Length of word string */
#define MAX_TAG_INFO    1000       /* Number of tags in a single document */

#define MAX_FILENAME     500       /* Length of filename */

#define MAX_LINE        5000       /* Length of one line in document/index file */


/********************/
/* Global variables */
/********************/


/* Tagset information */
/*--------------------*/
char   Tagset[MAX_TAGSET][MAX_TAG_STRING];
                     /* Tagset to annotate or extract */
                     /* This should not include attribute information */
                     /* For example, for <NAMEX type="location">,     */
                     /* only "NAMEX" will be the tag string           */

int    Tagset_N;              /* Number of Tagset information */


/* Document */
/*----------*/
unsigned char Document[MAX_DOC_LEN];    /* Storage for document */

struct s_docno {                 /* Document information from index_file      */
  char   s[MAX_DOCNO_STRING];    /*   Docno string                            */
  long   offset;                 /*   Offset of the information in index_file */
  int    mark;                   /*   If processed 1, unprocessed 0           */
} Docno[MAX_DOCNO];              /* element is refferd by ID */

int  Docno_N;                    /* Number of document information */


/* Tag information for each document */
/*-----------------------------------*/

struct s_tag_info {                /* Tag information */
  int  start;                      /*   start offset in the document */     
  int end;                         /*   end offset in the document   */
  char start_tag[MAX_TAG_STRING];  /*   start tag string             */
  char end_tag[MAX_TAG_STRING];    /*   end tag string               */
} Tag_info[MAX_TAG_INFO];          /* This data should be the order of start */


int Tag_info_N;                    /* Number of Tag_info                */


/* File  */
/*-------*/
char   Index_filename[MAX_FILENAME];   /* Index filename */
char   Doc_filename[MAX_FILENAME];     /* Document filename */
FILE  *Dfd;                            /* File descriptor for document file */
FILE  *Ifd;                            /* File descriptor for index file */
int    Dline;                          /* line number for document file */

/* Task */
/*------*/
int    Task;                       /* Task     */
                                   /*   0: extraction */
                                   /*   1: annotation */

/************************/
/* Function definitions */
/************************/
int main();
void read_arg();
void Usage();
void Error();
void init();
void read_index();
void extraction();
void annotation();
int read_document();
int find_docno();
int exist_tagset();
int tag_pair();
void read_tag_info();
void check_doc();
unsigned char *get_string();
int get_tagstring();
void Debug();

extern int strncmp();
extern int atoi();
extern long strtol();
extern int fprintf();
extern int fclose();
extern int fflush();
extern int fputc();
extern int fseek();


/********/
/* main */
/********/
int
main(argc,argv)
int argc;
char *argv;
{

    read_arg(argc,argv);

    init();

    read_index();

    if(Task==0){
	extraction();
    }else{
	annotation();
	check_doc();
    }

    return(0);
}


/*****************************************/
/* Read arguments                        */
/*  arguments must be                    */
/*   {-e|-a} index_filename doc_filename */
/*****************************************/
void
read_arg(argc,argv)
int argc;
char *argv[];
{
  if(argc!=4){
      Usage("Wrong number of arguments (%d)!!\n", argc);
  }

  if(strcmp(argv[1],"-e")==0){      
      Task = 0;
  }else if(strcmp(argv[1],"-a")==0){
      Task = 1;
  }else{
      Usage("Specify the task\n");
  }

  strcpy(Index_filename,argv[2]);
  strcpy(Doc_filename,argv[3]);
}


/***************************/
/* Print usage information */
/***************************/
void
Usage(s)
char *s;
{
  fprintf(stderr,s);
  fprintf(stderr,"Usage: tea {-a|-e} index_file doc_file\n");
  fprintf(stderr,"   Annotate tag to documents or extract tag from documents.\n");
  fprintf(stderr,"  See README file for more detail.            \n");
  fprintf(stderr,"                                              \n");
  fprintf(stderr,"Option:                                       \n");
  fprintf(stderr,"   a : Annotation mode                        \n");
  fprintf(stderr,"   e : Extraction mode                        \n");

  exit(1);
}


/********************************/
/* Print error message and exit */
/********************************/
void
Error(s,arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8)
char *s, *arg1, *arg2, *arg3, *arg4, *arg5, *arg6, *arg7, *arg8;
{
  fprintf(stderr,s,arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8);
  fprintf(stderr,"Sometime, index_file or output_file is helpful for\n");
  fprintf(stderr,"finding the bug\n");
  exit(1);
}


/*************************/
/* Global initialization */
/*************************/
void
init()
{

    if((Dfd=fopen(Doc_filename,"r"))==NULL){
	Error("Can't open document file (%s)\n",Doc_filename);
    }

    Dline = 0;
}


/*******************/
/* Read index file */
/*******************/
void
read_index()
{
  int line;
  unsigned char buff[MAX_LINE];
  int current_docno;

  if((Ifd=fopen(Index_filename,"r"))==NULL){
    Error("Can't open index file (%s)\n",Index_filename);
  }

  Tagset_N = 0;
  Docno_N = 0;
  current_docno = -1;

  for(line=1;fgets(buff,MAX_LINE,Ifd)!=NULL;line++){

  /* Comment */
  /*---------*/
    if(buff[0]=='#' || strlen(buff) < 3){

      continue;

  /* TAGSET */
  /*--------*/
    }else if(strncmp(buff,"TAGSET ",7)==0){

	get_string(Tagset[Tagset_N++],buff+7);

  /* DOCNO */
  /*-------*/
    }else if(strncmp(buff,"DOCNO ",6)==0 ||
	     strncmp(buff,"DOCID ",6)==0){

      if(Task == 0){
	Error("DOCNO defined in extraction task (index_file line=%d)\n",line);
      }

      get_string(Docno[Docno_N].s,buff+6);
      Docno[Docno_N].offset = ftell(Ifd);
      Docno[Docno_N].mark = 0;
      current_docno = Docno_N;

      Docno_N++;
      
  /* TAG INFORMATION */
  /*-----------------*/
    }else if(strncmp(buff,"@ ",2)==0){

      if(current_docno == -1){
	Error("Don't specify docno before tag information (%d)\n",line);
      }

  /* Other keyword line (ERROR) */
  /*----------------------------*/
    }else{

      Error("Unknown keyword in index file (%d)\n",line);

    }
  }

  fclose(Ifd);
}


/*******************/
/* Extraction MAIN */
/*******************/
void
extraction()
{
    int offset, len;
    int i;
    int start, end;
    unsigned char start_tag[MAX_TAG_STRING], end_tag[MAX_TAG_STRING];
    unsigned char word[MAX_WORD_STRING], *p;
    unsigned char docno[MAX_DOCNO_STRING];

    if((Ifd=fopen(Index_filename,"a"))==NULL){
	Error("Can't open index file (%s)\n",Index_filename);
    }

    while(read_document(docno)>=0){

	fprintf(Ifd,"DOCNO %s\n",docno);
	fflush(Ifd);
	len = strlen(Document);
	offset = 0;

	for(i=0;i<len;){
      
	    /* Find tag to be stored */
	    /*-----------------------*/
	    if(Document[i]=='<' && exist_tagset(Document+i+1)==1){

		/* Find start position */
		/*---------------------*/
		start = offset;
		i += get_tagstring(start_tag,Document+i);
		p = word;

		/* Find end position (Don't allow overlap, nesting) */
		/*--------------------------------------------------*/
		while(Document[i]!='\0'){

		    /* Find an end tag */
		    if(Document[i]=='<' && Document[i+1]=='/'){
			i += get_tagstring(end_tag,Document+i);

			/* If this is the paired end tag */
			if(tag_pair(start_tag,end_tag)==1){
			    break;
			} else {
			    Error("Unpaired tagset in <%s>\n",docno);
			}

		    /* Find a new start tag (Error) */
		    } else if(Document[i]=='<' && Document[i-1]!='\\'){
			
			Error("Unexpected start tag in <%s>\n",docno);

		    /* OTHERS */
		    } else {

			if(Document[i]>0x80 && Document[i+1]>0x80){
			    offset++;
			    *p++ = Document[i++];
			    *p++ = Document[i++];
			} else {
			    *p++ = Document[i++];
			}
		    }
		}    


		/* We found the end tag */
		/*----------------------*/
		*p = '\0';
		end = offset;
		fprintf(Ifd,"@ %d %d %s %s %s\n",start,end,start_tag,end_tag,word);
		fflush(Ifd);
		
	    /* Find unexpected end tag */
	    } else if(Document[i]=='<' && Document[i+1]=='/' &&
		      exist_tagset(Document+i+2)==1){

		Error("Unexpected end tag in <%s>\n",docno);

	    /* OTHERS */
	    } else if(Document[i]>0x80 && Document[i+1]>0x80){
		offset++;
		i+=2;

	    } else {
		i++;
	    }
	} 
    }

    fclose(Ifd);
}


void
annotation()
{
    unsigned char docno[MAX_DOCNO_STRING];
    int id;
    int offset, i;
    int len;
    int tag_id;

    if((Ifd=fopen(Index_filename,"r"))==NULL){
	Error("Can't open index file (%s)\n",Index_filename);
    }

  /* While there is document to annotate */
  /*-------------------------------------*/
    while((id=read_document(docno))>=0){

      /* Check and mark the document ID */
      /*--------------------------------*/
	if(Docno[id].mark==1){
	    Error("Document appeared more than once (%s)\n",Docno[id].s);
	}
	Docno[id].mark=1;

      /* Read tag information from index file */
      /*--------------------------------------*/
	read_tag_info(id);

      /* Create tagged document */
      /*------------------------*/
	len = strlen(Document);
	for(offset=0,i=0;i<len;){

          /* Write ASCII character */
	    if(Document[i]<0x80){

		fputc(Document[i++],stdout);

          /* Write Japanese character */
	    }else{

              /* start tag information */
		for(tag_id=0;tag_id<Tag_info_N;tag_id++){
		    if(Tag_info[tag_id].start==offset){
			fprintf(stdout,Tag_info[tag_id].start_tag);
		    }
		}

              /* character information */
		fputc(Document[i++],stdout);
		fputc(Document[i++],stdout);
		offset++;

              /* end tag information */
		for(tag_id=Tag_info_N-1;tag_id>=0;tag_id--){
		    if(Tag_info[tag_id].end==offset){
			fprintf(stdout,Tag_info[tag_id].end_tag);
		    }
		}
	    }
	}
    }
}


/*****************************************************************/
/* read_document(docno)                                          */
/*  In extraction task, read next document from document file.   */
/*  In annotation task, read next document which matches to one  */
/*                      of the docno in Docno[].s.               */
/*  The function returns document ID, if it finds a document,    */
/*  otherwise returns 0.  Set document Number into docno.        */
/*****************************************************************/
int
read_document(docno)
unsigned char *docno;
{
    unsigned char buff[MAX_LINE];
    unsigned char *p;
    unsigned char *ret;
    int i, j;
    int id;

    id = 0;

AGAIN:

    p = Document;

    while((ret=fgets(buff,MAX_LINE,Dfd))!=NULL){

	Dline++;
	if(strncmp(buff,"<DOC>",5)==0){
	    strcpy(p,buff);
	    p += strlen(buff);
	    break;
	}

    }

    if(ret==NULL){
	return(-1);
    }

    while((ret=fgets(buff,MAX_LINE,Dfd))!=NULL){

	Dline++;

      /* Read Document number */
      /*----------------------*/
	if(strncmp(buff,"<DOCNO>",7)==0 ||
	   strncmp(buff,"<DOCID>",7)==0){
	    for(i=7;isspace(buff[i]);i++);
	    if(buff[i]=='<' || buff[i]=='\0'){
		Error("No document number in document file (%d)\n",Dline);
	    }

	    for(j=0;!isspace(buff[i]) && buff[i]!='<' && buff[i]!='\0';){
		docno[j++]=buff[i++];
	    }
	    docno[j]='\0';
	  
	    if(Task==1 && (id=find_docno(docno))==-1){
		goto AGAIN;
	    }
	}

      /* Copy buff to Document */
      /*-----------------------*/
	strcpy(p,buff);
	p += strlen(buff);
      
    /* End of document */
    /*-----------------*/
	if(strncmp(buff,"</DOC>",6)==0){
	    break;
	}
    }

    if(ret==NULL){
	return(-1);
    }

    return(id);
}


/*************************************/
/* find_docno(docno)                 */
/*  if there is docno in Docno list, */
/*  return id, otherwise -1.         */
/*************************************/
int
find_docno(docno)
unsigned char *docno;
{
    int i;

    for(i=0;i<Docno_N;i++){

	if(strcmp(Docno[i].s,docno)==0){
	    return(i);
	}
    }

    return(-1);
}


/*******************************************/
/* exist_tagset(tag)                       */
/*  if there is tag in Tagset list,        */
/*  return 1, otherwise 0.                 */
/*  tag should be terminated by a space or */
/*  by '>'.                                */
/*******************************************/
int
exist_tagset(tag)
unsigned char *tag;
{
    int i;
    int len;

    for(i=0;i<Tagset_N;i++){
	len = strlen(Tagset[i]);

	if(strncmp(tag,Tagset[i],len)==0 &&
	   (isspace(tag[len]) || tag[len]=='>')){
	    return(1);
	}
    }

    return(0);
}

/***************************************/
/* tag_pair(start_tag,end_tag)         */
/*   If start_tag and end_tag is legal */
/*   pair, return 1, otherwise 0.      */
/*   Both of them should be surrounded */
/*   by '<' and '>'.                   */
/***************************************/
int
tag_pair(start_tag0,end_tag0)
unsigned char *start_tag0, *end_tag0;
{
    unsigned char start_tag[MAX_TAG_STRING];
    unsigned char end_tag[MAX_TAG_STRING];
    unsigned char *t, *s;

    for(s=start_tag0+1, t=start_tag;
	*s != '\0' && *s != '>' && !isspace(*s);s++,t++){
	*t = *s;
    }
    *t = '\0';

    for(s=end_tag0+2, t=end_tag;
	*s != '\0' && *s != '>' && !isspace(*s);s++,t++){
	*t = *s;
    }
    *t = '\0';

    if(strcmp(start_tag,end_tag)==0){
	return(1);
    } else {
	return(0);
    }
}
  

/************************************************/
/* Read tag information for document of (id)    */
/* from index file and store them into Tag_info */
/************************************************/
void
read_tag_info(id)
int id;
{
  unsigned char buff[MAX_LINE], *p;
  struct s_tag_info tag_info;
  int i, j;

  Tag_info_N = 0;

  fseek(Ifd,Docno[id].offset,0);

  while(fgets(buff,MAX_LINE,Ifd)!=NULL){

    if(buff[0]!='@'){
      break;
    }

    tag_info.start = (int)strtol(buff+1,&p,10);
    tag_info.end   = (int)strtol(p,&p,10);

    if(tag_info.start>=tag_info.end){
	fprintf(stderr,"Invalid tag information (end starts before start)\n");
	continue;
    }

    for(;isspace(*p);p++);
    for(i=0;*p!='>';p++,i++){
      tag_info.start_tag[i] = *p;
    }
    tag_info.start_tag[i++]=*p++;
    tag_info.start_tag[i]='\0';

    for(;isspace(*p);p++);
    for(i=0;*p!='>';p++,i++){
      tag_info.end_tag[i] = *p;
    }
    tag_info.end_tag[i++]=*p++;
    tag_info.end_tag[i]='\0';

  /* Store in the order of start, but if there are tags starting */
  /* the same position, the tag with larger end comes first      */
  /*-------------------------------------------------------------*/
    for(j=Tag_info_N;j>0;j--){
      if(tag_info.start>Tag_info[j-1].start || 
	 (tag_info.start==Tag_info[j-1].start &&
	  tag_info.end<Tag_info[j-1].end)){
	break;
      }
      Tag_info[j].start = Tag_info[j-1].start;
      Tag_info[j].end   = Tag_info[j-1].end;
      strcpy(Tag_info[j].start_tag,Tag_info[j-1].start_tag);
      strcpy(Tag_info[j].end_tag,Tag_info[j-1].end_tag);
    }

    Tag_info[j].start = tag_info.start;
    Tag_info[j].end   = tag_info.end;
    strcpy(Tag_info[j].start_tag,tag_info.start_tag);
    strcpy(Tag_info[j].end_tag,tag_info.end_tag);

    Tag_info_N++;
  }
}

void
check_doc()
{
  int i;

  for(i=0;i<Docno_N;i++){
    if(Docno[i].mark==0){
      Error("Tagged document is not created for (%s)\n",Docno[i].s);
    }
  }
}


/*********************************/
/* Copy non-space part of s to t */
/*********************************/
unsigned char *
get_string(t,s)
unsigned char *t, *s;
{
    unsigned char *p;

    p = t;

    for(;isspace(*s);s++);
    for(;!isspace(*s) && *s!='\0';s++,t++){
	*t = *s;
    }

    *t = '\0';

    return(p);
}


/********************************/
/* Copy tagstring of s to t     */
/* return the length of the tag */
/********************************/
int
get_tagstring(t,s)
unsigned char *t, *s;
{
    int i;

    for(i=0;*s!='>' && i<MAX_TAG_STRING;t++,s++,i++){
	*t = *s;
    }
    if(i>=MAX_TAG_STRING){
	Error("Tag name is too long\n");
    }
    *t++='>';
    *t='\0';

    return(i+1);
}

void
Debug(s,arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8)
char *s, *arg1, *arg2, *arg3, *arg4, *arg5, *arg6, *arg7, *arg8;
{
  if(DEBUG==1){
    fprintf(stderr,"# ");
    fprintf(stderr,s,arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8);
  }
}
